from geopy.geocoders import Nominatim, GoogleV3
from geopy.extra.rate_limiter import RateLimiter
import json

# Wrapper for geocoding api's, previous results are cached.
class GeoCode(object):
    """
    Wrapper for geopy's geocoding library,
    It simplifies the process of geocoding into two simple steps:
    1. Initialization (class instance creation)
    2. Geocoding (built in method within the class to transform an address into coordinates)

    This class has some perks such as a simple cache that can be used
    to avoid repetition of API calls into expensive services such as Google API, or just to improve speed.
    It has been tested with a cache size of around 6M unique addresses, which just work.
    Some optimization can be made, but if ain't broken don't fix it.
    (unless for some strange reason this becomes like the backbone of the internet, from which I'm sure 
    people far more versed coding will have their eyes bleed out first, and then try to improve on this c:)
    """
    def __init__(self, method:str = "GoogleV3", domain:str = "maps.google.cl", api_key:str = None, cache_file:str = None, min_delay:float = None):
        """ 
        Class constructor:
        Method -> string: Current options are: GoogleV3, Nominatim
        domain -> string: Address on which requests will be sent.
        api_key -> string: Needed for GoogleV3 API.
        cache_file -> string: File where requests and responses can be saved.
                              Default: adddress_cache.json on current directory. 
        min_delay -> float: Time interval between requests
                              By default it's 0.02 seconds (50 requests/second) on Google's API and 1 second on Nominatim.
        """
        self.method = method
        self.domain = domain
        self.api_key = api_key
        self.min_delay_seconds = None
        if min_delay is not None:
            self.min_delay_seconds = min_delay
        self.geolocator = self.set_geolocator()
        self.geocoder = RateLimiter(self.geolocator.geocode, min_delay_seconds=self.min_delay_seconds)
        
        # Cache file
        if cache_file is None:
            self.cache_file = "address_cache.json"
        else: 
            self.cache_file = cache_file
        
        self.cache = dict()
        
        # Cache loading
        try:
            self.load_cache()
        except:
            print("Creating cache file: "+self.cache_file)
            self.save_cache()
        
    def set_geolocator(self):
        """
        set_geolocator: Generates a geolocator based on Google Maps or Nominatim request type API
        returns a geopy geocoder.
        """
        if self.method == "GoogleV3":
            geolocator = GoogleV3(user_agent="LT", 
                                  domain = self.domain, 
                                  api_key = self.api_key)
            if self.min_delay_seconds is None:
                self.min_delay_seconds = 0.02 # Maximum of 50 requests per second.
        else:
            geolocator = Nominatim(user_agent="LT")
            if self.min_delay_seconds is None:
                self.min_delay_seconds = 1
        return geolocator
    
    # Obtains the saved address -> coordinates cache
    def get_cache(self):
        """
        get_cache corresponds to the cache attribute getter.
        returns a dictionary with the cache contents.
        """
        return self.cache
    
    # Loads a cache file.
    def load_cache(self):
        """
        load_cache corresponds to the cache attribute loader.
        It loads a previously saved cache file into memory. 
        """
        with open(self.cache_file, 'r') as f:
            self.cache = json.load(f)
    
    # Saves the cache in ram into the hard drive.
    def save_cache(self, verbose = True):
        """
        save_cache corresponds to the cache attribute dump.
        It saves the memory cache contents into a permanent location. 
        """
        if verbose:
            print("Saved in cache")
        with open(self.cache_file, 'w') as f:
            json.dump(self.cache, f)
            
    # Geocode transforms the address into coordinates using the selected geolocation engine.
    # When a request was already queried, it searches it in the cache, to save resources.
    def geocode(self, address, debug = False, save = False):
        if address not in self.cache:
            try:
                location = self.geocoder(address).raw
            except:
                location = "error"
                return location
            if debug: 
                return location
            # Address, coordinates in a json way
            location_coordinates = dict()
            if self.method == "GoogleV3":
                #in_rm = (location["formatted_address"].find("Metropolitana") != -1)
                location_coordinates["formatted_address"] = location["formatted_address"]
                location_coordinates["location"] = location["geometry"]["location"]
                location_coordinates["is_valid"] = (location["address_components"][0]["types"][0] == 'street_number') #& (in_rm) 
            else:
                location_coordinates["formatted_address"] = location["display_name"]
                coordinates = dict()
                coordinates["lat"] = location["lat"]
                coordinates["long"] = location["lon"]
                location_coordinates["location"] = coordinates
            self.cache[address] = location_coordinates
            if save:
                self.save_cache()
            
        return self.cache[address]




from easygeo import GeoCode
import geopy

def test_constructor():
    # Nominatim constructor
    g = GeoCode(method="Nominatim", domain="nominatim.openstreetmap.org", cache_file="./test/cache.json", min_delay = 2)
    assert(g.cache==dict())
    assert(g.api_key == None)
    assert(g.cache_file == "./test/cache.json")
    assert(g.domain == "nominatim.openstreetmap.org")
    assert(g.geocoder.__class__ == geopy.extra.rate_limiter.RateLimiter)
    assert(g.geolocator.__class__ == geopy.geocoders.nominatim.Nominatim)
    assert(g.min_delay_seconds == 2)
    assert(g.method == "Nominatim")
    # GoogleV3 constructor
    g2 = GeoCode(method="GoogleV3", api_key="aaa", cache_file="./test/cache_g.json")
    assert(g2.method == "GoogleV3")
    assert(g2.domain == "maps.google.cl")
    assert(g2.api_key == "aaa")
    assert(g2.cache_file == "./test/cache_g.json")
    assert(g2.min_delay_seconds == 0.02)
    # Initializing geocode with GoogleV3 api, will result in error.
    try:
        g3 = GeoCode()
    except:
        pass

def test_geolocator():
    g = GeoCode(method="Nominatim", domain="nominatim.openstreetmap.org", cache_file="./test/cache.json", min_delay = 2)
    geocoded_example = g.geocode("Beauchef 850, Santiago, RM, Chile")

    assert(list(geocoded_example.keys()) == ['formatted_address', "location"])
    assert(geocoded_example['formatted_address'] == "Escuela de Ingeniería Universidad de Chile, 850, Beauchef, Barrio República, Santiago, Provincia de Santiago, Región Metropolitana de Santiago, 8370456, Chile")
    assert(list(geocoded_example['location'].keys()) == ["lat", "long"])

def test_cache_load():
    g = GeoCode(method="Nominatim", domain="nominatim.openstreetmap.org", cache_file="./test/cache_test.json", min_delay = 2)
    geocoded_example = g.geocode("Beauchef 851, Santiago, RM, Chile", save = True)
    # New instance
    g2 = GeoCode(method="Nominatim", domain="nominatim.openstreetmap.org", cache_file="./test/cache_test.json", min_delay = 2)
    #g.load_cache()
    assert(len(g2.cache) == 1)
    geocoded_example = g.geocode("Beauchef 851, Santiago, RM, Chile")
    assert(len(g.cache) == 1)
    g.cache_file = "./test/cache.json"
    old_cache = g.cache 
    g.load_cache()
    assert(g.cache != old_cache)


#def 


from setuptools import setup

with open ("README.md", "r") as fh:
        long_description = fh.read()

setup(
    name="easygeo", 
    version="0.0.2",
    author="Joaquin Cubelli",
    author_email="jcubellidl@gmail.com",
    url="https://github.com/CubelliJ/easyGeo",
    description="Simple library that allows to geocode using Nominatim or Google V3 API",
    py_modules=["easygeo"],
    package_dir={"": "src"},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        "Operating System :: MacOS"
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires = [
        "geopy",
        "pandas"
    ],
    extra_require = {
        "dev": [
            "pytest>=7.0",
            "twine>4.0"
        ],
    },
)